import asyncio
import argparse


parser = argparse.ArgumentParser(description='fan commandline client')
parser.add_argument('cmd', nargs='*', help='command to execute')


async def run_cmd(cmd):
    from fan.sync import get_context
    ctx = get_context()
    with ctx:
        full_cmd = 'ctx.rpc.{}'.format(cmd)
        print('eval: {}'.format(full_cmd))
        print(eval(full_cmd))


def main():
    args = parser.parse_args()
    print('ARGS: {}'.format(args))

    loop = asyncio.get_event_loop()
    task = asyncio.ensure_future(run_cmd(args.cmd[0]))
    loop.run_until_complete(task)



from fan.utils import _make_key


def test_make_key():
    class A:
        pass
    args = (1, ('bla', ))
    kwargs = {'a': A()}

    key1 = _make_key(args, kwargs)
    key2 = _make_key(args, kwargs)
    assert key1 == key2, 'Keys must be the same'



from unittest import TestCase
from fan.lifecycle import Supervisor, Lifecycle, Spec, MaxRestarts, State


class Worker:
    def __init__(self):
        self.lifecycle = Lifecycle(self, 'on_start', 'terminate')
        self.started = False
        self.terminated = False

    def on_start(self):
        self.started = True

    def on_stop(self):
        self.lifecycle.terminate()

    def terminate(self, reason):
        self.started = False
        self.terminated = True


class DamagedWorker(Worker):
    def new_on_start(self):
        self.started = True

    def on_start(self):
        DamagedWorker.on_start = DamagedWorker.new_on_start
        raise Exception('damaged worker')


class BrokenWorkerError(Exception):
    pass


class BrokenWorker(Worker):
    def on_start(self):
        raise BrokenWorkerError


class FanTestSupervisor(Supervisor):

    def __init__(self):
        self.specs = [Spec(Worker)]
        super().__init__(self.specs)

    @property
    def state(self):
        return self.lifecycle.state


class LifecycleTest(TestCase):

    @property
    def s(self):
        if hasattr(self, '_s'):
            return self._s
        self._s = FanTestSupervisor()
        self._s.lifecycle.on_start()
        return self._s

    def test_01_simple_start(self):
        self.assertEqual(self.s.state, State.RUNNING)
        self.assertEqual(len(self.s.instances), 1)

    def test_02_simple_add(self):
        worker = list(self.s.instances)[0]
        self.assertEqual(worker.lifecycle.state, State.RUNNING)

    def test_03_worker_restart(self):
        worker = list(self.s.instances)[0]
        self.assertEqual(worker.lifecycle.state, State.RUNNING)
        worker.lifecycle.terminate()
        self.assertEqual(worker.lifecycle.state, State.TERMINATED)

        worker = list(self.s.instances)[0]
        self.assertEqual(worker.lifecycle.state, State.RUNNING)

    def test_04_damaged_restart(self):
        self.s.add_child(Spec(DamagedWorker))
        workers = list(self.s.instances)
        self.assertEqual(len(workers), 2)
        worker2 = workers[1]
        self.assertEqual(self.s.state, State.RUNNING)
        self.assertEqual(worker2.lifecycle.state, State.RUNNING)

        self.s.lifecycle.terminate()
        self.assertEqual(self.s.lifecycle.state, State.TERMINATED)
        self.assertEqual(len(self.s.instances), 0)
        self.assertEqual(workers[0].lifecycle.state, State.TERMINATED)
        self.assertEqual(workers[1].lifecycle.state, State.TERMINATED)

    def test_05_add_broken_worker(self):
        self.assertRaises(MaxRestarts, self.s.add_child, Spec(BrokenWorker))
        self.assertEqual(self.s.lifecycle.state, State.TERMINATED)

    def test_06_broken_supervisor(self):
        s = Supervisor([Spec(BrokenWorker)])
        self.assertRaises(MaxRestarts, s.lifecycle.on_start)
        self.assertEqual(list(s.instances), [])
        self.assertEqual(s.lifecycle.state, State.TERMINATED)

    def test_07_supervisor_of_supervisor(self):
        sup_spec = Spec(Supervisor, [Spec(BrokenWorker)])
        s = Supervisor([sup_spec])
        self.assertRaises(MaxRestarts, s.lifecycle.on_start)
        self.assertEqual(s.lifecycle.state, State.TERMINATED)

    def test_08_nested_add(self):
        sup_spec = Spec(Supervisor, [Spec(DamagedWorker)])
        s = Supervisor([sup_spec])
        s.lifecycle.on_start()
        self.assertEqual(s.lifecycle.state, State.RUNNING)

        s2 = list(s.instances)[0]
        self.assertEqual(s2.lifecycle.state, State.RUNNING)

        self.assertRaises(MaxRestarts, s2.add_child, Spec(BrokenWorker))
        self.assertEqual(s2.lifecycle.state, State.TERMINATED)
        self.assertEqual(s.lifecycle.state, State.TERMINATED)



import json
import logging
import re

from kazoo.client import KazooClient

from fan.discovery import RemoteDiscovery
from fan.remote import ProxyEndpoint


VSN_RE = re.compile(r'^\d+(\.\d+){,2}$')
CONN_TIMEOUT = 5


class KazooWrapper:
    """
    Wrapper that ensure that we connect to zookeeper only when need it.
    Prevents connection on error request.
    """
    log = logging.getLogger('KazooWrapper')

    def __init__(self, chroot=None, *args, **kwargs):
        self.zk = KazooClient(*args, **kwargs)
        self.chroot = chroot

    @property
    def _started(self):
        return self.zk.connected

    def start(self, timeout=CONN_TIMEOUT):
        self.log.debug('Connect')
        self.zk.start(timeout=timeout)
        if self.chroot:
            self.zk.ensure_path(self.chroot)
            self.zk.chroot = self.chroot

    def stop(self):
        if self._started:
            self.zk.stop()

    def __getattr__(self, name):
        if not self._started:
            self.start()
        return getattr(self.zk, name)


class KazooDiscovery(RemoteDiscovery):
    log = logging.getLogger('KazooDiscovery')
    timeout = 5

    def __init__(self, zk_path, chroot='/'):
        super().__init__()
        self.zk = KazooWrapper(chroot=chroot, hosts=zk_path, timeout=self.timeout)

    def on_start(self):
        self.zk.start(timeout=self.timeout)

    def on_stop(self):
        self.zk.stop()

    def find_endpoint(self, service_tuple, version_filter):
        path = ['/endpoints'] + list(service_tuple)
        path = '/'.join(path)

        if not self.zk.exists(path):
            return
        childs = self.zk.get_children(path)
        if len(childs) == 0:
            return
        childs = [x for x in childs if VSN_RE.match(x)]
        version = sorted(childs)[-1]
        vpath = '{}/{}'.format(path, version)
        configs = self.zk.get_children(vpath)
        # keep only config_ paths
        configs = filter(lambda x: x.startswith('config_'), configs)
        return self.create_endpoint(service_tuple, vpath, configs)

    def get_transport_class(self, name):
        return self.transport_classes[name]

    def create_endpoint(self, name, path, configs):
        # TODO: pass all configs
        configs = sorted(configs)
        assert len(configs), (name, path, configs)
        config = configs[-1]
        dpath = '{}/{}'.format(path, config)
        data = self.zk.get(dpath)[0].decode('utf8')
        self.log.debug('GET DATA: {!r}'.format(data))
        params = json.loads(data)
        return ProxyEndpoint(self, name, params)



import asyncio
import os
import re
import sys
import yaml

from fan.contrib.sync_helper import SyncHelper
from tipsi_tools.tipsi_logging import setup_logger


R = re.compile('%\{(.*?)\}')
P = '%{{{}}}'.format
D = os.path.dirname(__file__)


def env(s):
    for k in R.findall(s):
        s = s.replace(P(k), os.environ.get(k, 'NO_VARIABLE_{}'.format(k)))
    return s


def set_fields(dct):
    for k, v in dct.items():
        if not v and k in os.environ:
            dct[k] = os.environ[k]


def traverse(obj):
    t = type(obj)
    if t == dict:
        return {traverse(k): traverse(v) for k, v in obj.items()}
    elif t == list:
        return list(map(traverse, obj))
    elif t == str:
        return env(obj)
    return obj


def setup_local_ip():
    if not os.path.exists('/.dockerenv'):
        return
    with open('/etc/hosts') as f:
        lines = f.readlines()
    local_ip = lines[-1].strip().split()[0]
    os.environ['LOCAL_IP'] = local_ip


def main():
    if 'LOG_DIR' not in os.environ:
        os.makedirs('logs', exists_ok=True)
        os.environ['LOG_DIR'] = './logs'
    setup_logger('fan')

    args = sys.argv[1:]
    assert len(args) == 1, 'Give yaml config name'
    zk_config = os.environ.get('ZK_HOST')
    zk_chroot = os.environ.get('ZK_CHROOT', '/')
    setup_local_ip()
    conf = traverse(yaml.load(open(args[0])))
    print('Call with conf: {}'.format(conf))
    loop = asyncio.get_event_loop()
    h = SyncHelper(zk_config, zk_chroot, conf)
    loop.create_task(h.on_start())
    print('run loop')
    loop.run_forever()


if __name__ == '__main__':
    main()



import logging
from functools import wraps
from types import FunctionType
from typing import List, Set, Any  # noqa


class ServiceGroup:
    '''
    define services
    define routes
    encapsulate discovery and config
    start and stop services
    '''
    services = []  # type: List[Any]

    def __init__(self, discovery):
        self.log = logging.getLogger(self.__class__.__name__)
        self.discovery = discovery
        self.instances = []

    def start(self):
        for service_class in self.services:
            service = service_class()
            service.on_start()
            self.instances.append(service)

    def stop(self):
        for service in self.instances:
            service.on_stop()

    def get_connections(self) -> Set[str]:
        connections = set()
        for srv_dict in self.services:
            for e_dict in srv_dict.get('endpoints', []):
                assert 'connection' in e_dict, '"connection" key must be defined in endpoint config'  # noqa
                connections.add(e_dict['connection'])
        return connections


class ServiceMeta(type):
    log = logging.getLogger('ServiceMeta')

    def run_asserts(name, attrs):
        if name == 'Service':
            return
        assert attrs['name'], 'No service name for {} given'.format(name)

    def process_rpc(svc, bases, attrs):
        if svc.__name__ == 'Service':
            return

        r = svc._rpc = {}

        for k, v in attrs.items():
            if hasattr(v, 'rpc'):
                fun = getattr(svc, v.__name__)
                r[v.rpc['name']] = fun.__name__

        for base in bases:
            if hasattr(base, '_rpc'):
                for k, v in base._rpc.items():
                    if k not in r:
                        r[k] = v
        r['_meta'] = {'name': tuple(svc.name.split('.'))}

    def __new__(cls, name, bases, attrs, **kwargs):
        cls.log.debug('cls: {} n: {} b: {} A: {}'.format(cls, name, bases, attrs, **kwargs))
        cls.run_asserts(name, attrs)
        obj = super().__new__(cls, name, bases, attrs, **kwargs)
        cls.process_rpc(obj, bases, attrs)
        return obj


class Service(metaclass=ServiceMeta):
    name = None  # type: str
    version = None  # type: tuple

    def __init__(self):
        self.log = logging.getLogger(self.name)

    def on_start(self):
        pass

    def on_stop(self):
        pass


def endpoint(name):
    """
    you can just wrap function with @endpoint and it set current function name as name
    or you can pass name - @endpoint('custom_name')
    """
    if type(name) == FunctionType:
        name_string = name.__name__
        return endpoint(name_string)(name)

    def set_rpc(func):
        func.rpc = {
            'name': name or func.__name__
        }

    def wrapped(func):
        set_rpc(func)
        @wraps(func)
        def ret(svc, *args, **kwargs):
            assert isinstance(svc, Service), '{} is not subclass of Service'.format(svc)
            return func(svc, *args, **kwargs)
        return ret
    return wrapped



from fan.lifecycle import (Terminated, MaxRestarts, State,
                           Spec, Lifecycle, Supervisor)


def async_terminate_on_error(fun):
    async def wrapped(self, *args, **kwargs):
        try:
            res = await fun(self, *args, **kwargs)
            return res
        except Exception as e:
            self.log.debug('Catch exception: {!r}'.format(e))
            try:
                if hasattr(self, 'lifecycle'):
                    await self.lifecycle.terminate(e)
                else:
                    await self.terminate(e)
            except Exception as term_e:
                self.log.exception('during terminate')
            raise e
    return wrapped


class AIOSpec(Spec):
    async def run(self, obj):
        await obj.lifecycle.on_start()
        return obj


class AIOLifecycle(Lifecycle):

    @async_terminate_on_error
    async def on_start(self):
        assert self.state == State.INITIALIZED, self.state
        self.state = State.STARTING
        if self.start_method:
            await getattr(self.obj, self.start_method)()
        self.state = State.RUNNING

    async def safe_call(self, fun, *args, **kwargs):
        try:
            return await fun(*args, **kwargs)
        except Exception as e:
            self.log.info('during call {} [{} {}]'.format(fun, args, kwargs))

    async def terminate(self, reason=Terminated):
        if self.state >= State.STOPPING:
            return
        self.state = State.STOPPING
        self.log.info('Terminate reason: {!r} {}'.format(reason, self.obj))
        for link in self.links:
            await self.safe_call(link.terminate, reason, self.obj)
        for watch in self.watches:
            await self.safe_call(watch.notify, reason, self.obj)
        self.state = State.STOPPED
        if self.terminate_method:
            try:
                self.state = State.TERMINATING
                await self.safe_call(getattr(self.obj, self.terminate_method), reason)
            except Exception as e:
                self.log.exception('during terminate: {}.{}'.format(self.obj,
                                                                    self.terminate_method))
        self.state = State.TERMINATED

    @async_terminate_on_error
    async def notify(self, reason, obj):
        await getattr(self.obj, self.notify_method)(reason, obj)


class AIOSupervisor(Supervisor):
    def __init__(self, specs=[]):
        super().__init__(specs)
        self.lifecycle = AIOLifecycle(self, 'on_start', 'terminate', 'notify')

    @async_terminate_on_error
    async def notify(self, reason, obj):
        if self.lifecycle.state > State.RUNNING:
            return
        try:
            spec = self.instances.pop(obj)
        except KeyError as e:
            self.log.exception('Get notify without from obj not in self.instances')
            raise e
        await self.start_child(spec)

    async def add_child(self, spec):
        assert self.lifecycle.state == State.RUNNING, \
          'Supervisor should be started before add child {}'.format(spec)
        self.specs.append(spec)
        await self.start_child(spec)

    @async_terminate_on_error
    async def start_child(self, spec):
        while True:
            try:
                child = spec.get_obj()
                await spec.run(child)
                self.lifecycle.watch(child)
                self.instances[child] = spec
                return child
            except MaxRestarts as e:
                raise e
            except Exception:
                self.log.info('Restart worker: {}'.format(spec))

    async def on_start(self):
        self.log.debug('Start supervisor')
        assert self.lifecycle.state == State.STARTING, "don't call on_start directly"
        for spec in self.specs:
            await self.start_child(spec)

    async def terminate(self, reason):
        assert self.lifecycle.state == State.TERMINATING, "don't call terminate directly"
        self.log.info('Supervisor terminate reason: {}'.format(reason))
        while True:
            try:
                obj, spec = self.instances.popitem()
                await obj.lifecycle.terminate(reason)
            except KeyError:
                break



from unittest import TestCase

from basictracer import BasicTracer
from basictracer.recorder import InMemoryRecorder
from fan.context import Context
from fan.discovery import (SimpleDictDiscovery,
                           LocalDiscovery, CompositeDiscovery)
from fan.service import ServiceGroup, Service, endpoint
from fan.process import Process
from fan.remote import LocalEndpoint, RemoteEndpoint
from fan.tests import DummyTracer, DummyTransport


class FanTestServiceGroup(ServiceGroup):
    def __init__(self, discovery):
        super().__init__(discovery)
        self.endpoints = []

    def start(self):
        for data in self.services:
            service = data['service']()
            service.on_start()

            self.instances.append(service)
            for ep_conf in data['endpoints']:
                ep = ep_conf['endpoint'](self.discovery, service, ep_conf['params'])
                ep.on_start()
                self.endpoints.append(ep)
                self.discovery.register(ep)


class DummyServiceGroup(FanTestServiceGroup):
    services = [{'service': DummyTracer,
                 'endpoints': [{'endpoint': RemoteEndpoint,
                                'params': {'id': 1, 'transport': 'dummy'}}]}]


class ChainedEchoService(Service):
    name = 'chained_echo'

    @endpoint('echo')
    def echo(self, ctx, word):
        return ctx.rpc.dummy_tracer.echo(word, 0)[0]


class ChainedServiceGroup(FanTestServiceGroup):
    services = [{'service': ChainedEchoService,
                 'endpoints': [{'endpoint': RemoteEndpoint,
                                'params': {'id': 2, 'transport': 'dummy'}}]}]


class FanTestProcess(Process):
    def create_context(self):
        return Context(self.discovery)


class FanTestRemoteDiscovery(SimpleDictDiscovery):
    def __init__(self, conf):
        super().__init__(conf)
        self.conf = conf
        self.remote = None

    def link(self, other_remote):
        if not self.remote:
            self.remote = other_remote
            self.remote.link(self)

    def find_local_endpoint(self, service_name):
        self.log.debug('Remote Lookup: {} {}'.format(service_name, self.cached_endpoints))
        if service_name in self.cached_endpoints:
            return self.cached_endpoints[service_name]
        assert False, 'Cannot find: {} {}'.format(service_name, self.cached_endpoints)

    def find_remote_endpoint(self, service_name, version_filter):
        pass


class FanTestLocalDiscovery(LocalDiscovery):
    transports = {'dummy': DummyTransport}

    def get_transport_class(self, name):
        return self.transports[name]


class ProcessTestCase(TestCase):
    def setUp(self):
        self.recorder = InMemoryRecorder()
        d = CompositeDiscovery(LocalDiscovery(), SimpleDictDiscovery({}))
        d.tracer = BasicTracer(self.recorder)

        self.process = FanTestProcess(d)

    def test_call(self):
        context = self.process.create_context()

        context.discovery.register(LocalEndpoint(DummyTracer()))
        with context:
            response = context.rpc.dummy_tracer.echo('hello', 7)

        self.assertEquals(len(self.recorder.get_spans()), 9)
        assert response == ('hello', 0), response


class MultiProcessTestCase(TestCase):
    def setUp(self):
        self.recorder = InMemoryRecorder()

        self.remote = FanTestRemoteDiscovery({})

        self.p1 = self.create_process(self.recorder, ChainedServiceGroup)
        self.p2 = self.create_process(self.recorder, DummyServiceGroup)
        self.p1.start()
        self.p2.start()

    def create_process(self, recorder, sg):
        discovery = CompositeDiscovery(FanTestLocalDiscovery(), self.remote)
        discovery.tracer = BasicTracer(recorder)
        proc = FanTestProcess(discovery)
        proc.service_groups = [sg]
        return proc

    def test_call(self):
        context = self.p1.create_context()
        with context:
            result = context.rpc.chained_echo.echo('hello')
        self.assertEqual(result, 'hello')
        self.assertEquals(len(self.recorder.get_spans()), 3)



import asyncio
from types import CoroutineType

from basictracer.context import SpanContext

from fan.context import Context
from fan.remote import ProxyEndpoint, Transport, RemoteEndpoint


class AIOTransport(Transport):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.loop = asyncio.get_event_loop()
        self.remote = isinstance(self.endpoint, RemoteEndpoint)
        self.responses = {}

    def terminate(self, reason):
        for v in self.responses.values():
            v.set_exception(reason)

    async def wait_or_terminate(self, coro):
        f = asyncio.ensure_future(coro)
        ret = await asyncio.wait([f, self.terminate_future], return_when=asyncio.FIRST_COMPLETED)
        return await ret[0].pop()

    async def rpc_call(self, name, ctx, *args, **kwargs):
        # span_id, trace_id, sampled, baggage, with_baggage_item

        c = ctx.span.context
        context_headers = {'span_id': c.span_id,
                           'trace_id': c.trace_id,
                           'sampled': c.sampled}
        self.log.debug('CTX: {}'.format(context_headers))
        msg = {'context_headers': context_headers,
               'method': name,
               'args': args,
               'kwargs': kwargs}
        f = self.responses[str(c.span_id)] = asyncio.Future()
        resp = await self.rpc_inner_call(msg, f)
        return resp['response']

    async def read_loop(self, *args, **kwargs):
        try:
            while not self.stopped:
                msg = await self.inner_read_message(*args, **kwargs)
                assert type(msg) == dict, 'Msg is: {!r}'.format(msg)
                ctx_headers = msg['context_headers']
                if self.remote:
                    parent_ctx = SpanContext(**ctx_headers)
                    method = msg['method']
                    ctx = Context(self.discovery, self.endpoint.service,
                                  parent_ctx, method)
                    self.log.debug('CTX: {}'.format(ctx.span.context.trace_id))
                    call_args = msg.get('args', ())
                    call_kwargs = msg.get('kwargs', {})
                    hc = self.handle_call(method, ctx, *call_args, **call_kwargs)
                    if isinstance(hc, CoroutineType):
                        resp = await hc
                    else:
                        resp = hc
                    self.log.debug('Remote send resp ==> {}'.format(resp))
                    response = {'context_headers': msg['context_headers'],
                                'method': msg['method'],
                                'response': resp}
                    await self.remote_send_response(msg, response)
                else:
                    self.log.debug('Proxy return resp: {}'.format(msg))
                    self.proxy_send_response(msg)
        except asyncio.CancelledError:
            pass
        except Exception as e:
            self.log.exception('In read loop')
            self.terminate(e)

    async def remote_send_response(self, request, response):
        raise NotImplementedError

    async def inner_read_message(self, *args, **kwargs):
        raise NotImplementedError

    def proxy_send_response(self, msg):
        span_id = str(msg['context_headers']['span_id'])
        self.responses.pop(span_id).set_result(msg)

    async def rpc_inner_call(msg, future):
        """
        implemente here concrete interaction with pub/sub transport level
        """
        raise NotImplementedError

    async def on_start(self):
        raise NotImplementedError

    async def on_stop(self):
        raise NotImplementedError


class AIOQueueBasedTransport:
    """
    Mixin to build queue based RPC
    """
    async def on_start(self):
        await self.sub_prepare()
        await self.pub_prepare()

    async def on_stop(self):
        await self.pub_stop()
        await self.sub_stop()

    def sub_prepare(self):
        raise NotImplementedError

    def sub_stop(self):
        raise NotImplementedError

    def pub_prepare(self):
        raise NotImplementedError

    def pub_stop(self):
        raise NotImplementedError


class AIOProxyEndpoint(ProxyEndpoint):

    async def perform_call(self, ctx, method_name, *args, **kwargs):
        if not self.transport.started:
            await self.transport.on_start()
        return await self.transport.rpc_call(method_name, ctx, *args, **kwargs)

    async def on_start(self):
        await self.transport.on_start()

    async def on_stop(self):
        await self.transport.on_stop()



# TODO: Move sanic service helper to separate module

import asyncio
import logging
import os
from typing import Type

from sanic import Sanic

from fan.asynchronous import get_discovery, AsyncContext
from fan.contrib.sync_helper import SyncHelper


class SanicRegister:
    log = logging.getLogger('EndpointRegister')

    def __init__(self):
        self.sync_helper = None

    def get_local_ip(self):
        if not os.path.exists('/.dockerenv'):
            return '127.0.0.1'
        with open('/etc/hosts') as f:
            lines = f.readlines()
        return lines[-1].strip().split()[0]

    def __init__(self, name, port, transport='http', version='1.0.0'):
        self.zk_config = os.environ.get('ZK_HOST', 'zk')
        self.zk_chroot = os.environ.get('ZK_CHROOT', '/')
        self.service = {
            'methods': [],
            'name': name,
            'host': self.get_local_ip(),
            'port': port,
            'transport': transport,
            'version': version,
        }
        self.methods = []

    def add(self, name, url, method='GET', content_type='application/json'):
        self.service['methods'].append({
            'name': name,
            'method': method,
            'url': url,
            'content_type': content_type,
        })

    async def register(self, app=None, loop=None):
        assert app or loop, 'Either app nor loop must be provided as keyword argument'
        conf = {
            'services': [self.service]
        }
        self.sync_helper = SyncHelper(self.zk_config, self.zk_chroot, conf)
        self.loop = app.loop if app else loop
        await self.sync_helper.on_start()

    async def stop(self):
        await self.sync_helper.on_stop()


class AbstractTaskWorker:
    def __init__(self, loop):
        self.idle_timeout = 1
        self.loop = loop
        self.stopping = False
        self.task = self.loop.create_task(self.worker_loop())

    async def worker_loop(self):
        while not self.stopping:
            await asyncio.sleep(self.idle_timeout, loop=self.loop)

    def stop(self):
        self.stopping = True
        if not self.task.done():
            self.task.cancel()

    @classmethod
    def task(cls, app=None, loop=None):
        assert app or loop, 'Either app nor loop must be defined'
        return cls(app.loop if app else loop)

    @classmethod
    async def deferred_task(cls, app):
        # This method is used with sanic_app.add_task to retrieve loop after app started.
        return cls(app.loop)


class SanicServiceHelper:
    DEFAULT_BAGGAGE_HEADERS = {
        'HTTP_INSTALLATION_ID': 'INSTALLATION_ID',
        'HTTP_SESSION_ID': 'SESSION_ID',
    }

    def __init__(self, name, host, port, **kwargs):
        self._host = host
        self._port = port
        self.sanic_server = None
        self.app = Sanic(name)
        self.fan_reg = SanicRegister(name, port=self._port)
        self.task_classes = []
        self.workers = []
        self.baggage_headers = kwargs.get('baggage_headers', self.DEFAULT_BAGGAGE_HEADERS)

    def add_endpoint(self, handler, name, url, method, **kwargs):  # TODO: url arguments
        self.app.add_route(handler, url, methods=[method])
        self.fan_reg.add(name=name, url=url, method=method, **kwargs)

    def add_task(self, task_class: Type[AbstractTaskWorker]):
        self.task_classes.append(task_class)

    def add_exception_handler(self, handler):
        self.app.error_handler.add(Exception, handler)

    def prepare_app(self, loop):
        def fan_exception_handler(request, exception):
            return _inner_handler(request, exception)

        def _inner_handler(request, exception):
            request['fan_exc_info'] = exception

            # Temporally remove fan_handler from handlers
            # TODO: Think about better way to override error handling
            new_handlers = []
            for exc, hdl in self.app.error_handler.handlers:
                if hdl != fan_exception_handler:
                    new_handlers.append((exc, hdl))
            self.app.error_handler.handlers[:] = new_handlers

            result = self.app.error_handler.response(request=request, exception=exception)

            self.add_exception_handler(fan_exception_handler)
            return result

        self.add_exception_handler(fan_exception_handler)

        @self.app.middleware('request')
        async def fan_sanic_on_request(request):
            if request.get('fan_ctx'):
                return

            discovery = await get_discovery(name=self.app.name, loop=loop)
            tracer = discovery.tracer

            span_context = tracer.extract('http', request.headers)
            name = '{} {}'.format(request.method, request.path)
            if span_context:
                 ctx = AsyncContext(discovery, None, span_context, name)
            else:
                # TODO: Extract same code for django and sanic
                baggage = {}
                for k, v in self.baggage_headers.items():
                    h = request.headers.get(k)
                    if h is not None:
                        baggage[v] = h
                ctx = AsyncContext(discovery, None, None, name, baggage)

            await ctx.__aenter__()
            request['fan_ctx'] = ctx

        @self.app.middleware('response')
        async def fan_sanic_on_response(request, response):
            ctx = request.get('fan_ctx')
            if ctx:
                exc_info = request.get('fan_exc_info')
                await ctx.__aexit__(exc_info and type(exc_info), exc_info and str(exc_info), None)

    def run(self, **kwargs):
        loop = kwargs.get('loop') or asyncio.get_event_loop()
        asyncio.ensure_future(self.async_run(loop=loop, **kwargs), loop=loop)
        loop.run_forever()

    async def async_run(self, loop=None, **kwargs):
        loop = loop or asyncio.get_event_loop()
        for task_class in self.task_classes:
            self.workers.append(task_class.task(loop=loop))
        server = self.app.create_server(host=self._host, port=self._port, **kwargs)
        self.prepare_app(loop=loop)
        self.sanic_server = asyncio.ensure_future(server, loop=loop)
        await self.fan_reg.register(loop=loop)

    async def stop(self):
        await self.fan_reg.stop()
        if not self.sanic_server.done():
            self.sanic_server.cancel()
        for worker in self.workers:
            worker.stop()



import asyncio
import os

from tipsi_tools.testing.aio import AIOTestCase

from fan.contrib.aio.discovery import ZKDiscovery
from fan.tests import TEST_TIMEOUT


class StubEndpoint:
    name = 'this.is.stub.endpoint'
    version = '1.0.0'
    config = '{"stub": "config"}'


class TestZK(AIOTestCase):
    async def setUp(self):
        self.remote = ZKDiscovery(os.environ.get('ZK_HOST', 'zk:2181'), chroot='/test_fan')
        await asyncio.wait_for(self.remote.on_start(), TEST_TIMEOUT)

    async def test_init(self):
        pass

    async def test_register_function(self):
        data = b'{"test": "data"}'
        await self.remote.recursive_create('/endpoints/test/sub1/sub2/1.0.0', data)
        base_path = '/endpoints/test/sub1/sub2/1.0.0'
        names = await self.remote.zk.get_children(base_path)
        out = await self.remote.zk.get_data('{}/{}'.format(base_path, names[0]))
        self.assertEqual(data, out)
        await self.remote.zk.deleteall('/endpoints/test')

    async def test_register_endpoint(self):
        ep = StubEndpoint()
        await self.remote.register(ep)

    async def test_register_barrier(self):
        ep = StubEndpoint()
        zk = self.remote.zk
        barrier_name = '{}/{}/barrier'.format(ep.name, ep.version)
        barrier = zk.recipes.Barrier(barrier_name)
        await barrier.create()
        fut = asyncio.ensure_future(self.remote.register(ep))
        try:
            await asyncio.wait_for(fut, 0.12)
        except asyncio.TimeoutError:
            pass
        await barrier.lift()
        await self.remote.register(ep)



import asyncio

import aiohttp
import pytest
from sanic import response

from fan.asynchronous import get_context
from fan.contrib.sync_helper.sanic_helpers import AbstractTaskWorker, SanicServiceHelper


async def ping(request):
    return response.json(['pong'])


@pytest.fixture
def ft_task_is_running():
    yield asyncio.Future()


@pytest.fixture
def fun_test_task(ft_task_is_running):
    class FunTestTask(AbstractTaskWorker):
        async def worker_loop(self):
            while not self.stopping:
                await asyncio.sleep(0.1)
                ft_task_is_running.set_result(True)
                await asyncio.sleep(self.idle_timeout)

    return FunTestTask


@pytest.fixture
def fan_test_service_port(unused_tcp_port_factory):
    yield unused_tcp_port_factory()


@pytest.fixture
def fan_test_service_url(fan_test_service_port):
    yield f'http://127.0.0.1:{fan_test_service_port}'


@pytest.fixture
def fan_test_service(fan_test_service_port, fun_test_task):
    service = SanicServiceHelper('test_service', host='0.0.0.0', port=fan_test_service_port)
    service.add_endpoint(ping, name='ping', url='/ping/', method='GET')
    service.add_task(fun_test_task)
    yield service


@pytest.fixture
async def running_service(event_loop, fan_test_service):
    await fan_test_service.async_run(loop=event_loop)
    yield fan_test_service
    await fan_test_service.stop()


@pytest.fixture
async def fan_async_context(event_loop):
    async with await get_context(loop=event_loop) as ctx:
        yield ctx
        await ctx.stop()


@pytest.mark.asyncio
async def test_server_is_running(running_service, fan_test_service_url, ft_task_is_running):
    await asyncio.wait_for(ft_task_is_running, timeout=1)
    assert ft_task_is_running.result(), 'Task is not running'


@pytest.mark.asyncio
async def test_server_http_ok(running_service, fan_test_service_url, ft_task_is_running):
    conn = aiohttp.TCPConnector()
    async with aiohttp.ClientSession(connector=conn) as session:
        async with session.get(f'{fan_test_service_url}/ping/') as response:
            response_data = await response.json()
            assert 'pong' in response_data, response_data


@pytest.mark.asyncio
async def test_get_async_context_two_times(event_loop):
    async with await get_context(loop=event_loop) as ctx:
        await ctx.stop()

    async with await get_context(loop=event_loop) as ctx:
        await ctx.stop()


@pytest.mark.asyncio
async def test_rpc_async_call(event_loop, running_service, fan_test_service_url, fan_async_context):
    resp = await fan_async_context.rpc.test_service.ping()
    assert 'pong' in resp, resp


@pytest.mark.asyncio
async def test_call_async_ctx_second_type(event_loop, running_service, fan_test_service_url,
                                          fan_async_context):
    # Test that different event_loop is cached in cache discovery
    resp = await fan_async_context.rpc.test_service.ping()
    assert 'pong' in resp, resp



import logging

from fan.remote import LocalEndpoint, Endpoint, ProxyEndpoint, RemoteEndpoint


def path_set(obj, path, what):
    curr = obj
    for name in path[:-1]:
        if name not in curr:
            curr[name] = {}
        curr = curr[name]
    curr[path[-1]] = what


def path_unset(obj, path):
    curr = obj
    for name in path[:-1]:
        if name not in path:
            return
        curr = curr[name]
    del curr[path[-1]]


def path_get(obj, path):
    curr = obj
    for name in path:
        if name not in path:
            return
        curr = curr[name]
    return curr


class LocalDiscovery:
    """
    Service types:
        singleton - only a single service allowed
        round_robin - any worker can handle request
        shard - only a concrete worker can handle request
    """
    def __init__(self):
        self.log = logging.getLogger(self.__class__.__name__)
        self.cached_endpoints = {}

    def register(self, endpoint: Endpoint):
        if isinstance(endpoint, LocalEndpoint):
            service = endpoint.service
            path = tuple(service.name.split('.'))
            self.cached_endpoints[path] = endpoint
        else:
            path = tuple(endpoint.name)
            self.cached_endpoints[path] = endpoint

    def find_endpoint(self, service_name, version_filter=None):
        if service_name in self.cached_endpoints:
            return self.cached_endpoints[service_name]


class RemoteDiscovery:
    '''
    Doesn't contain any objects, only hierarchical data with primitive datatypes
    May cache data, but should invalidate cached in distributed deployments
    '''
    def __init__(self):
        self.log = logging.getLogger(self.__class__.__name__)
        self.watchers = {}

    def register(self, endpoint):
        raise NotImplementedError

    def find_endpoint(self, service_name, version_filter):
        return self.find_remote_endpoint(service_name, version_filter)

    def find_remote_endpoint(self, service_name, version_filter):
        raise NotImplementedError

    def watch(self, path, callback):
        raise NotImplementedError

    def unwatch(self, path, callback):
        raise NotImplementedError


def version_filter(constraints):
    def filter_versions(versions):
        if len(versions):
            # TODO: add actual filter
            return versions[0]
    return filter_versions


class CompositeDiscovery:
    def __init__(self, local, remote):
        self.log = logging.getLogger(self.__class__.__name__)
        self.local = local
        self.remote = remote

    def find_endpoint(self, name, version_filter=[]):
        local = self.local.find_endpoint(name, version_filter)
        if local:
            return local
        proxy_cfg = self.remote.find_endpoint(name, version_filter)
        if proxy_cfg:
            ep = self.create_proxy(name, proxy_cfg)
            self.local.register(ep)
            return ep

    def create_proxy(self, name, proxy_cfg):
        return ProxyEndpoint(self, name, proxy_cfg)

    def register(self, endpoint):
        self.local.register(endpoint)
        if isinstance(endpoint, RemoteEndpoint):
            self.remote.register(endpoint)

    # TODO: transport info should go from process
    def get_transport_class(self, name):
        return self.local.get_transport_class(name)


class SimpleDictDiscovery(RemoteDiscovery):
    def __init__(self, conf):
        super().__init__()
        self.data = conf

    def register(self, endpoint):
        path, data = endpoint.service.name.split('.'), endpoint.remote_params
        path_set(self.data, path, data)

    def find_endpoint(self, path, version_filter):
        return path_get(self.data, path)

    def watch(self, path, cb):
        pass

    def unwatch(self, path, cb):
        pass



'''
Shortlived sync helpers. Primary target is creating short-lived fast context with get_context
'''
import logging
import os
import socket
from functools import wraps

import requests
from basictracer import BasicTracer
from basictracer.recorder import SpanRecorder
from py_zipkin.thrift import (annotation_list_builder, create_endpoint,
                              binary_annotation_list_builder, create_span, span_to_bytes,
                              encode_bytes_list)
from py_zipkin.zipkin import ZipkinAttrs

from fan.context import Context
from fan.contrib.kazoo.discovery import KazooDiscovery
from fan.transport import HTTPTransport, HTTPPropagator, DjangoPropagator

discovery = None
tracer = None
ZIPKIN = os.environ.get('ZIPKIN')


def http_transport(encoded_span):
    # copypasted from: https://github.com/Yelp/py_zipkin#transport
    # The collector expects a thrift-encoded list of spans. Instead of
    # decoding and re-encoding the already thrift-encoded message, we can just
    # add header bytes that specify that what follows is a list of length 1.
    requests.post(
        'http://{}/api/v1/spans'.format(ZIPKIN),
        data=encoded_span,
        headers={'Content-Type': 'application/x-thrift'},
    )


log = logging.getLogger('fan.sync')


def zipkin_log_span(span_id, parent_span_id, trace_id, span_name, annotations, binary_annotations,
                    timestamp_s, duration_s, **kwargs):
    span = create_span(span_id, parent_span_id, trace_id, span_name, annotations,
                       binary_annotations, timestamp_s, duration_s)
    http_transport(encode_bytes_list([span_to_bytes(span)]))


class BaseFanRecorder(SpanRecorder):
    _endpoint = None
    my_ip = socket.gethostbyname(socket.gethostname())

    def __init__(self, name, send_to_zipkin):
        super().__init__()
        self.name = name
        self.send_to_zipkin = send_to_zipkin

    @property
    def endpoint_info(self):
        return {
            'port': 80,
            'service_name': self.name,
            'ipv4': self.my_ip,
        }

    @property
    def endpoint(self):
        if not self._endpoint:
            self._endpoint = create_endpoint(**self.endpoint_info)
        return self._endpoint

    def zipkin_span_params(self, span):
        span_params = self.log_span_params(span)

        timing = {
            'ss': span.start_time,
            'sr': span.start_time + span.duration
        }
        annotations = annotation_list_builder(timing, self.endpoint)
        zipkin_attrs = ZipkinAttrs(
            trace_id=span_params['trace_id'],
            span_id=span_params['span_id'],
            parent_span_id=span_params['parent_span_id'],
            flags=span_params['flags'],
            is_sampled=span_params['is_sampled']
        )

        return {
            **span_params,
            'zipkin_attrs': zipkin_attrs,
            'annotations': annotations,
            'binary_annotations': binary_annotation_list_builder({
                **span.context.baggage,
                **span.tags
            }, self.endpoint),
        }

    def log_span_params(self, span):
        ctx = span.context
        return {
            'trace_id': hex(ctx.trace_id)[2:],
            'span_id': hex(ctx.span_id)[2:],
            'parent_span_id': span.parent_id and hex(span.parent_id)[2:],
            'service_name': self.name,
            'span_name': span.operation_name or 'no_name',

            'flags': 0,  # for ZipkinAttrs
            'is_sampled': True,  # for ZipkinAttrs
            'endpoint': self.endpoint_info,  # for annotation

            'timestamp_s': span.start_time,
            'duration_s': span.duration,
            'span_context_baggage': span.context.baggage,
            'span_context_tags': span.tags,
        }


class FanRecorder(BaseFanRecorder):
    def record_span(self, span):
        if self.send_to_zipkin:
            zipkin_log_span(**self.zipkin_span_params(span))

        log.info('Log span: {}'.format(self.log_span_params(span)))

        return super().record_span(span)


def get_tracer(name=None):
    global tracer
    if tracer:
        return tracer

    recorder = FanRecorder(name or 'no_name', send_to_zipkin=ZIPKIN)
    tracer = BasicTracer(recorder)
    return tracer


def cache_discovery(fun):
    @wraps(fun)
    def wrapped(*args, **kwargs):
        global discovery
        if discovery:
            return discovery
        discovery = fun(*args, **kwargs)
        return discovery

    return wrapped


@cache_discovery
def get_discovery(is_django=False, name=None):
    discovery = KazooDiscovery(os.environ.get('ZK_HOST', 'zk'), os.environ.get('ZK_CHROOT', '/'))
    discovery.transport_classes = {
        'http': HTTPTransport,
    }

    if not name and is_django:
        name = 'django'

    discovery.tracer = get_tracer(name)
    if is_django:
        discovery.tracer.register_propagator('http', DjangoPropagator())
    else:
        discovery.tracer.register_propagator('http', HTTPPropagator())
    return discovery


def get_context(name=None, service_name=None):
    if not service_name and name:
        service_name = name
    return Context(get_discovery(is_django=False, name=service_name), name=name)



import logging


class Transport:
    def __init__(self, discovery, endpoint, params):
        self.log = logging.getLogger(self.__class__.__name__)
        self.discovery = discovery
        self.params = params
        self.endpoint = endpoint
        self.started = False
        self.stopped = False

    def on_start(self):
        self.started = True

    def on_stop(self):
        self.stopped = True

    def rpc_call(self, method, ctx, *args, **kwargs):
        raise NotImplementedError

    def handle_call(self, method, ctx, *args, **kwargs):
        return getattr(self.endpoint, method)(ctx, *args, **kwargs)


class Endpoint:
    name = None  # type: str

    def __init__(self, service):
        self.log = logging.getLogger(self.__class__.__name__)
        self.service = service
        self.started = False
        self.stopped = False

    def on_start(self):
        self.started = True

    def on_stop(self):
        self.stopped = True


class ProxyEndpoint(Endpoint):
    def __init__(self, discovery, name, params):
        self.log = logging.getLogger(self.__class__.__name__)
        self.name = name
        self.params = params
        self.discovery = discovery
        transportClass = discovery.get_transport_class(params['transport'])
        self.transport = transportClass(discovery, self, params)

    def perform_call(self, ctx, method_name, *args, **kwargs):
        if not self.transport.started:
            self.transport.on_start()
        # ctx.span.operation_name = method_name
        return self.transport.rpc_call(method_name, ctx, *args, **kwargs)

    def on_start(self):
        self.transport.on_start()

    def on_stop(self):
        return self.transport.on_stop()


class LocalEndpoint(Endpoint):
    """
    For in-process communications only
    """
    def __init__(self, service):
        super().__init__(service)
        self.name = service.name.split('.')
        self.log = logging.getLogger(self.__class__.__name__)

    def perform_call(self, ctx, method_name, *args, **kwargs):
        return getattr(self.service, self.service._rpc[method_name])(ctx, *args, **kwargs)


class RemoteEndpoint(LocalEndpoint):
    """
    Listen some remote transport for incoming messages
    """

    def __init__(self, discovery, service, params):
        self.service = service
        self.params = params
        self.remote_params = params
        if 'transport' in params:
            self.transportClass = discovery.get_transport_class(params['transport'])
        else:
            self.transportClass = Transport
        self.transport = self.transportClass(discovery, self, params)

    @property
    def name(self):
        return self.service.name

    def on_start(self):
        self.transport.on_start()

    def on_stop(self):
        return self.transport.on_stop()

    def __getattr__(self, name):
        return getattr(self.service, self.service._rpc[name])



from fan.rpc import RPC
from fan.service import Service


class Context:
    def __init__(self, discovery, service=None, parent=None, name=None, baggage=None):
        self.discovery = discovery
        if service:
            assert isinstance(service, Service), service
        self.service = service
        self.parent = parent

        if isinstance(parent, Context):
            parent_context = parent.span.context
        else:
            parent_context = parent
        self.span = discovery.tracer.start_span(child_of=parent_context, operation_name=name)
        if baggage:
            for k, v in baggage.items():
                self.span.set_baggage_item(k, v)
        self._entered = False

    def create_child_context(self, name=None):
        return self.__class__(self.discovery, self.service, self, name)

    @property
    def rpc(self):
        '''
        if you generate totally new context, you must enter it before use:
        >>> with new_ctx:
               # now you may call new_ctx.rpc
               new_ctx.rpc.some_endpoint.ping()
        for most services it would be started when accepting incoming request in RemoteEndpoint,
        so you should not do that in services
        '''
        assert self._entered, 'You must enter context before call .rpc'
        return RPC(self)

    def pre_call(self):
        pass

    def post_call(self):
        self.span.finish()

    def __enter__(self, *args):
        self._entered = True
        self.pre_call()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            self.span.set_tag('error', '{} {}'.format(exc_type, exc_val))
        self.post_call()


class AsyncContext(Context):
    @property
    def rpc(self):
        """
        if you generate totally new context, you must enter it before use:
        >>> async with new_ctx:
               # now you may call new_ctx.rpc
               await new_ctx.rpc.some_endpoint.ping()
        for most services it would be started when accepting incoming request in RemoteEndpoint,
        so you should not do that in services
        """
        assert self._entered, 'You must enter context before call .rpc'
        return RPC(self, async_caller=True)

    async def pre_call(self):
        pass

    async def post_call(self):
        await self.span.finish()

    async def __aenter__(self, *args):
        self._entered = True
        await self.pre_call()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            self.span.set_tag('error', '{} {}'.format(exc_type, exc_val))
        await self.post_call()

    async def stop(self):
        await self.discovery.zk.session.close()



from unittest import TestCase
from fan.remote import RemoteEndpoint, Transport
from fan.discovery import SimpleDictDiscovery, CompositeDiscovery, LocalDiscovery
from fan.service import Service, endpoint


class DummyTransport(Transport):
    def on_start(self):
        super().on_start()
        self.test_started = True


class DummyDiscovery(SimpleDictDiscovery):
    pass


class FanTestLocalDiscovery(LocalDiscovery):
    transports = {'dummy': DummyTransport}

    def get_transport_class(self, name):
        return self.transports[name]


class DummyService(Service):
    name = 'dummy'

    @endpoint
    def ping(self, ctx):
        return 'pong'


class DummyRemoteEndpoint(RemoteEndpoint):
    transportClass = DummyTransport


class TransportCase(TestCase):
    def setUp(self):
        self.dict_discovery = DummyDiscovery({})
        self.discovery = CompositeDiscovery(FanTestLocalDiscovery(), self.dict_discovery)
        self.svc = DummyService()
        self.svc.on_start()

    def test_remote_register(self):
        ep = DummyRemoteEndpoint(self.discovery, self.svc, {'transport': 'dummy'})
        ep.on_start()
        self.discovery.register(ep)

        assert ep.transport.test_started
        l = self.discovery.local
        r = self.discovery.remote
        assert l.cached_endpoints[('dummy',)] == ep
        assert r.data['dummy'] == {'transport': 'dummy'}



from sanic.response import json

from fan.contrib.sync_helper.sanic_helpers import SanicServiceHelper


async def ping(request):
    return json(['pong'])


service = SanicServiceHelper('service_two', host='127.0.0.1', port=31416)

service.add_endpoint(ping, name='ping', url='/ping/', method='GET')

service.run()



from sanic.response import text, json

from fan.contrib.sync_helper.sanic_helpers import SanicServiceHelper


async def f_ok(request):
    resp = await request['fan_ctx'].rpc.service_two.ping()
    return json([f"OK 2: {resp}"])


async def f_error(request):
    resp = await request['fan_ctx'].rpc.service_two.ping()
    1 / 0
    return text(f'OK 1: {resp}')


service = SanicServiceHelper('fan_test', host='127.0.0.1', port=31415)

service.add_endpoint(f_ok, name='ok', url='/ok/', method='GET')
service.add_endpoint(f_error, name='error', url='/error/', method='GET')

service.run()



from setuptools import setup, find_packages

with open('fan/__init__.py', 'r') as f:
    for line in f:
        if line.startswith('__version__'):
            version = line.strip().split('=')[1].strip(' \'"')
            break

setup(
    name='fan',
    packages=find_packages(),
    version=version,
    description='microservices kit',
    author='cybergrind',
    author_email='cybergrind@gmail.com',
    keywords=['rpc', 'microservices'],
    url='https://github.com/micro-fan/fan',
    classifiers=[],
    entry_points={
        'console_scripts': [
            'fan_register=fan.contrib.sync_helper.fan_register:main',
            'fan=fan.scripts.fan:main',
        ]
    },
    install_requires=[
        'aiozk>=0.15',
        'opentracing',
        'basictracer',
        'requests',
        'kazoo',
        'tipsi_tools[logging]>=1.49.0',
        'py-zipkin>=0.18.0,<1.0.0',
        'aiohttp',
        'sanic',
    ],
)



_split_object = object()


def _make_key(args, kwargs):
    key = args
    if kwargs:
        key += (_split_object, )  # split args kwargs with unique object
        for item in kwargs.items():
            key += item

    return key


def async_cache(fun):
    cache = {}

    async def _inner(*args, **kwargs):
        key = _make_key(args, kwargs)
        if key not in cache:
            result = await fun(*args, **kwargs)
            cache[key] = result

        return cache[key]

    return _inner



import os

import yaml
from tipsi_tools.testing.aio import AIOTestCase
from aiozk import ZKClient

from fan.contrib.sync_helper import SyncHelper


CURR_DIR = os.path.dirname(os.path.abspath(__file__))
CONF = yaml.load(open(os.path.join(CURR_DIR, 'conf.yaml')))


class TestSyncHelper(AIOTestCase):

    async def dump_tree(self, base='/'):
        out = list(await self.get_tree(base))
        print('Tree dump: {}'.format(out))
        return out

    @property
    def zk_path(self):
        return os.environ.get('ZK_HOST', 'zk')

    async def setUp(self):
        self.zk = ZKClient(self.zk_path)
        await self.zk.start()

    async def get_tree(self, curr='/'):
        out = [curr]
        childs = await self.zk.get_children(curr)
        for c in childs:
            # eliminate double slash: //root = '/'.join('/', 'root')
            if curr == '/':
                curr = ''
            out.append(await self.get_tree('/'.join([curr, c])))
        return out

    async def test_00_simple(self):
        conf = CONF
        s = SyncHelper(self.zk_path, '/', conf)
        await s.on_start()



"""
Short-lived async helpers. Primary target is creating short-lived fast context with get_context
"""

import asyncio
import logging
import os
import socket
import time

import aiohttp
from basictracer import BasicTracer
from basictracer.span import BasicSpan
from py_zipkin.thrift import create_span, span_to_bytes, encode_bytes_list

from fan.context import AsyncContext
from fan.contrib.aio.discovery import LazyAiozkDiscovery
from fan.sync import BaseFanRecorder
from fan.transport import AsyncHTTPTransport, HTTPPropagator
from fan.utils import async_cache

discovery = None
tracer = None
ZIPKIN = os.environ.get('ZIPKIN')
ZIPKIN_TIMEOUT = 10  # Timeout to send span to zipkin over http


async def async_http_transport(encoded_span):
    # Adapted from: https://github.com/Yelp/py_zipkin#transport
    # The collector expects a thrift-encoded list of spans. Instead of
    # decoding and re-encoding the already thrift-encoded message, we can just
    # add header bytes that specify that what follows is a list of length 1.
    timeout = aiohttp.ClientTimeout(total=ZIPKIN_TIMEOUT)
    try:
        async with aiohttp.ClientSession(timeout=timeout) as session:
            url = 'http://{}/api/v1/spans'.format(ZIPKIN)
            headers = {'Content-Type': 'application/x-thrift'}
            async with session.post(url, data=encoded_span, headers=headers) as resp:
                await resp.read()
    except asyncio.TimeoutError as e:
        log.warning('Failed to send span to zipkin. Timeout occurred.')


MY_IP = socket.gethostbyname(socket.gethostname())
log = logging.getLogger('fan.async')


async def async_zipkin_log_span(span_id, parent_span_id, trace_id, span_name, annotations,
                                binary_annotations, timestamp_s, duration_s, **kwargs):
    span = create_span(span_id, parent_span_id, trace_id, span_name, annotations,
                       binary_annotations, timestamp_s, duration_s)
    await async_http_transport(encode_bytes_list([span_to_bytes(span)]))


class AsyncFanRecorder(BaseFanRecorder):
    async def record_span(self, span):
        if self.send_to_zipkin:
            await async_zipkin_log_span(**self.zipkin_span_params(span))

        log.info('Log span: {}'.format(self.log_span_params(span)))

        return super().record_span(span)


class AsyncSpan(BasicSpan):
    async def finish(self, finish_time=None):
        finish = time.time() if finish_time is None else finish_time
        self.duration = finish - self.start_time
        await self._tracer.record(self)


class AsyncTracer(BasicTracer):
    def start_span(self, operation_name=None, child_of=None, references=None, tags=None,
                   start_time=None):
        span = super().start_span(operation_name, child_of, references, tags, start_time)
        return AsyncSpan(self, span.operation_name, span._context, span.parent_id, span.tags,
                         span.start_time)

    async def record(self, span):
        await self.recorder.record_span(span)


def get_async_tracer(name=None):
    global tracer
    if tracer:
        return tracer

    recorder = AsyncFanRecorder(name or 'no_name', send_to_zipkin=ZIPKIN)
    tracer = AsyncTracer(recorder)
    return tracer


@async_cache
async def get_discovery(name=None, loop=None):
    zk_host = os.environ.get('ZK_HOST', 'zk')
    zk_chroot = os.environ.get('ZK_CHROOT', '/')
    discovery = LazyAiozkDiscovery(zk_host, zk_chroot, with_data_watcher=False, loop=loop)
    discovery.transport_classes = {
        'http': AsyncHTTPTransport,
    }

    discovery.tracer = get_async_tracer(name)
    discovery.tracer.register_propagator('http', HTTPPropagator())
    return discovery


async def get_context(name=None, service_name=None, loop=None):
    if not service_name and name:
        service_name = name
    discovery = await get_discovery(name=service_name, loop=loop)
    return AsyncContext(discovery, name=name)



import asyncio

from fan.asynchronous import get_context


async def run(loop):
    async with await get_context(name='service_client', loop=loop) as ctx:
        # Normal request example
        resp = await ctx.rpc.service_one.ok()
        print(resp)

        # Error example
        resp = await ctx.rpc.service_one.error()
        print(resp)


loop = asyncio.get_event_loop()
task = loop.create_task(run(loop=loop))
loop.run_until_complete(task)



import logging

from fan.exceptions import RPCException


class Caller:
    log = logging.getLogger('Caller')

    def __init__(self, parent_context, name, call_path=[]):
        self.parent_context = parent_context
        self.call_path = call_path + [name]

    def __getattr__(self, name):
        '''
        Make copy of self, usefull for caches, like:
        >>> app = ctx.rpc.app
        >>> author = app.author
        >>> author.list()
        >>> author.delete(id=1)
        '''
        return self.__class__(self.parent_context, name, self.call_path)

    def __call__(self, *args, **kwargs):
        self.log.debug('Generate call: {} {} {}'.format(self.call_path, args, kwargs))
        endpoint = self.parent_context.discovery.find_endpoint(
            tuple(self.call_path[:-1]), version_filter=None
        )
        self.log.debug('RPCEndpoint: {}'.format(endpoint))
        if endpoint:
            name = '.'.join(self.call_path)
            with self.parent_context.create_child_context(name=name) as ctx:
                method_name = self.call_path[-1]
                return endpoint.perform_call(ctx, method_name, *args, **kwargs)
        else:
            raise RPCException('No such endpoint')


class AsyncCaller(Caller):
    async def __call__(self, *args, **kwargs):
        self.log.debug('Generate call: {} {} {}'.format(self.call_path, args, kwargs))
        endpoint = await self.parent_context.discovery.find_endpoint(
            tuple(self.call_path[:-1]), version_filter=None
        )
        self.log.debug('RPCEndpoint: {}'.format(endpoint))
        if endpoint:
            name = '.'.join(self.call_path)
            async with self.parent_context.create_child_context(name=name) as ctx:
                method_name = self.call_path[-1]
                return await endpoint.perform_call(ctx, method_name, *args, **kwargs)
        else:
            raise RPCException('No such endpoint')


class RPC:
    def __init__(self, context, async_caller=False):
        self.context = context
        self.async_call = async_caller

    def __getattr__(self, name):
        return (AsyncCaller if self.async_call else Caller)(self.context, name)



import os

from fan.contrib.redis import RedisEndpoint, RedisTransport
from fan.contrib.aio.tests import AIOEndpointCase


class RedisCase(AIOEndpointCase):
    endpoint_class = RedisEndpoint
    transport_class = RedisTransport
    endpoint_params = {'host': os.environ.get('REDIS_HOST', 'redis'),
                       'port': 6379,
                       'transport': 'redis',
                       'queue': 'test'}

    async def test_call(self):
        await self._test_remote_call()

    async def test_remote_register(self):
        await self._test_remote_register()



from fan.service import Service, endpoint
from fan.remote import Transport, RemoteEndpoint


TEST_TIMEOUT = 5


class DummyService(Service):
    name = 'dummy'

    @endpoint('echo')
    def method(self, context, message):
        self.log.debug('SELF: {} MSG: {}'.format(self, message))
        return message


class NestedService(DummyService):
    name = 'nested.tree.dummy'


class DummyTracer(Service):
    '''
    Call itself given number of times
    '''
    name = 'dummy_tracer'

    @endpoint('echo')
    def echo(self, ctx, word, count):
        if count > 0:
            return ctx.rpc.dummy_tracer.echo(word, count-1)
        else:
            return word, 0


class DummyTransport(Transport):
    storage = {}  # type: dict

    def __init__(self, discovery, endpoint, params):
        super().__init__(discovery, endpoint, params)
        if isinstance(endpoint, RemoteEndpoint):
            self.storage[params['id']] = self

    def rpc_call(self, method, ctx, *args, **kwargs):
        self.log.debug('Storage: {}'.format(self.storage))
        remote_ep = self.storage[self.params['id']]
        return remote_ep.handle_call(method, ctx, *args, **kwargs)
